import datetime
import pandas as pd
import numpy as np
from ...api.raw import RawDataApi
from ..raw.raw_data_helper import RawDataHelper
from ....data.view.raw_models import QSFundStyleBox
from ....util.calculator_item import CalculatorBase

class OverseaFundStyleBox:

    DEFAULT_START_DATE = '2018-01-01'
    HK_CH_EXCHANGE_RATE = 0.8
    US_CH_EXCHANGE_RATE = 6.5

    def __init__(self, data_helper: RawDataHelper):
        self._data_helper = data_helper
        self._raw_api = RawDataApi()

    def init(self):
        self.fund_info = self._raw_api.get_over_sea_fund_info()
        self.fund_pos = self._raw_api.get_fund_hold_top10_pos()
        #ch_equity_list, us_equity_list, hk_equity_list = self.get_stock_list(self.fund_pos)
        self.ch_stock_fin = self._raw_api.get_em_stock_fin_fac()
        hm_stock_fin = self.process_fin_fac(self._raw_api.get_os_stock_fin_fac())
        self.us_stock_fin = hm_stock_fin[~hm_stock_fin.stock_id.str.contains('\.HK')]
        self.hk_stock_fin = hm_stock_fin[hm_stock_fin.stock_id.str.contains('\.HK')]

        # 股价
        self.ch_stock_price = self._raw_api.get_em_stock_price(start_date=self.DEFAULT_START_DATE, columns=['close']).set_index(['stock_id', 'datetime'])
        self.hk_stock_price = self._raw_api.get_em_hk_stock_price(start_date=self.DEFAULT_START_DATE).replace(0,np.nan).ffill()
        self.us_stock_price = self._raw_api.get_em_us_stock_price(start_date=self.DEFAULT_START_DATE).replace(0,np.nan).ffill()
        dic = {'CODES':'stock_id','DATES':'datetime','CLOSE':'close'}
        self.hk_stock_price = self.hk_stock_price.rename(columns=dic).pivot_table(index='datetime',columns='stock_id',values='close')
        self.us_stock_price = self.us_stock_price.rename(columns=dic).pivot_table(index='datetime',columns='stock_id',values='close')
        self.ch_stock_price = self.ch_stock_price.reset_index().pivot_table(index='datetime',columns='stock_id',values='close')

        # 市值
        self.cap = self.process_cap()

    def process_fin_fac(self, df):
        td = pd.to_datetime(df.report_date)
        td = [i.date() for i in td]
        df.report_date = td
        return df.rename(columns={'report_date':'datetime','stock_code':'stock_id'})

    def us_stock_index_map(self, df):
        df.columns = [i.split('.')[0]+' '+'US Equity' for i in df.columns]
        return df

    def hk_stock_index_map(self, df):
        df.columns = [i.split('.')[0].lstrip('0')+' '+'HK Equity' for i in df.columns]
        return df

    def ch_stock_index_map(self, df):
        df.columns = [i.split('.')[0]+' '+'CH Equity' for i in df.columns]
        return df
    
    #  overall value score 
    def process_cap(self):
         # 总股本 部分中概股总市值用abs 计算 abs 和 流通股有的差距100倍 比如PTR，不统一，这里选择直接用市值
        ch_total_share = self._raw_api.get_em_daily_info(start_date=self.DEFAULT_START_DATE, columns=['total_share']).pivot_table(index='datetime', columns='stock_id', values='total_share')
        hk_cap = self.hk_stock_fin.pivot_table(index='datetime', columns='stock_id', values='indicator_18').ffill().iloc[-1]
        us_cap = self.us_stock_fin.pivot_table(index='datetime', columns='stock_id', values='indicator_18').ffill().iloc[-1]

        self.ch_cap = pd.DataFrame(ch_total_share.iloc[-1] * self.ch_stock_price.iloc[-1])
        self.ch_cap.columns = ['cap']
        self.us_cap = pd.DataFrame(us_cap * self.US_CH_EXCHANGE_RATE)
        self.us_cap.columns = ['cap']
        self.hk_cap = pd.DataFrame(hk_cap * self.HK_CH_EXCHANGE_RATE)
        self.hk_cap.columns = ['cap']
        return pd.concat([self.ch_cap, self.us_cap, self.hk_cap])

    def process_ep(self):
        ch_ep = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='np_ttmrp').ffill().iloc[-1]
        ch_ep = pd.DataFrame(ch_ep / self.ch_cap.cap)
        ch_ep.columns = ['ep']
        us_ep = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_3').ffill().iloc[-1]
        us_ep = pd.DataFrame(us_ep / self.us_cap.cap * self.US_CH_EXCHANGE_RATE)
        us_ep.columns = ['ep']
        hk_ep = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_3').ffill().iloc[-1]
        hk_ep = pd.DataFrame(hk_ep / self.hk_cap.cap * self.HK_CH_EXCHANGE_RATE)
        hk_ep.columns = ['ep']
        return pd.concat([ch_ep,us_ep,hk_ep]).dropna().rank(pct=True) * 100

    def process_bp(self):
        ch_bp = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='balance_statement_74').ffill().iloc[-1]
        ch_bp = pd.DataFrame(ch_bp / self.ch_cap.cap)
        ch_bp.columns = ['bp']
        us_bp = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_4').ffill().iloc[-1] / self.us_stock_price.iloc[-1]
        us_bp = pd.DataFrame(us_bp)
        us_bp.columns = ['bp']
        hk_bp = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_4').ffill().iloc[-1] / self.hk_stock_price.iloc[-1]
        hk_bp = pd.DataFrame(hk_bp)
        hk_bp.columns = ['bp']
        return pd.concat([ch_bp,us_bp,hk_bp]).dropna().rank(pct=True) * 100

    def process_rp(self):
        ch_rp = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='or_ttmr').ffill().iloc[-1]
        ch_rp = pd.DataFrame(ch_rp / self.ch_cap.cap)
        ch_rp.columns = ['rp']
        us_rp = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_13').ffill().iloc[-1] / self.us_stock_price.iloc[-1]
        us_rp = pd.DataFrame(us_rp)
        us_rp.columns = ['rp']
        hk_rp = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_13').ffill().iloc[-1] / self.hk_stock_price.iloc[-1]
        hk_rp = pd.DataFrame(hk_rp)
        hk_rp.columns = ['rp']
        return pd.concat([ch_rp,us_rp,hk_rp]).dropna().rank(pct=True) * 100

    def process_cp(self):
        ch_cp = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='cfo_ttmr').ffill().iloc[-1]
        ch_cp = pd.DataFrame(ch_cp / self.ch_cap.cap)
        ch_cp.columns = ['cp']
        us_cp = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_11').ffill().iloc[-1] / self.us_stock_price.iloc[-1]
        us_cp = pd.DataFrame(us_cp)
        us_cp.columns = ['cp']
        hk_cp = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_11').ffill().iloc[-1] / self.hk_stock_price.iloc[-1]
        hk_cp = pd.DataFrame(hk_cp)
        hk_cp.columns = ['cp']
        return pd.concat([ch_cp,us_cp,hk_cp]).dropna().rank(pct=True) * 100
    
    # overall growth score 近似用每股数据的变化率做
    def process_er(self):
        ch_er = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='np_ttmrp').ffill().pct_change(1).mean().dropna()
        ch_er = pd.DataFrame(ch_er)
        ch_er.columns = ['er']
        us_er = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_3').ffill().pct_change(1).mean()
        us_er = pd.DataFrame(us_er)
        us_er.columns = ['er']
        hk_er = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_13').ffill().pct_change(1).mean()
        hk_er = pd.DataFrame(hk_er)
        hk_er.columns = ['er']
        return pd.concat([ch_er,us_er,hk_er]).dropna().rank(pct=True) * 100

    def process_br(self):
        ch_br = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='balance_statement_74').ffill().pct_change(1).mean().dropna()
        ch_br = pd.DataFrame(ch_br)
        ch_br.columns = ['br']
        us_br = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_4').ffill().pct_change(1).mean()
        us_br = pd.DataFrame(us_br)
        us_br.columns = ['br']
        hk_br = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_4').ffill().pct_change(1).mean()
        hk_br = pd.DataFrame(hk_br)
        hk_br.columns = ['br']
        return pd.concat([ch_br,us_br,hk_br]).dropna().rank(pct=True) * 100

    def process_rr(self):
        ch_rr = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='or_ttmr').ffill().pct_change(1).mean().dropna()
        ch_rr = pd.DataFrame(ch_rr / self.ch_cap.cap)
        ch_rr.columns = ['rr']
        us_rr = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_13').ffill().pct_change(1).mean().dropna()
        us_rr = pd.DataFrame(us_rr)
        us_rr.columns = ['rr']
        hk_rr = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_13').ffill().pct_change(1).mean().dropna()
        hk_rr = pd.DataFrame(hk_rr)
        hk_rr.columns = ['rr']
        return pd.concat([ch_rr,us_rr,hk_rr]).dropna().rank(pct=True) * 100

    def process_cr(self):
        ch_cr = self.ch_stock_fin.pivot(index='datetime', columns='stock_id', values='cfo_ttmr').ffill().pct_change(1).mean().dropna()
        ch_cr = pd.DataFrame(ch_cr / self.ch_cap.cap)
        ch_cr.columns = ['cr']
        us_cr = self.us_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_11').ffill().pct_change(1).mean().dropna()
        us_cr = pd.DataFrame(us_cr)
        us_cr.columns = ['cr']
        hk_cr = self.hk_stock_fin.pivot(index='datetime', columns='stock_id', values='indicator_11').ffill().pct_change(1).mean().dropna()
        hk_cr = pd.DataFrame(hk_cr)
        hk_cr.columns = ['cr']
        return pd.concat([ch_cr,us_cr,hk_cr]).dropna().rank(pct=True) * 100

    def calculate_stock_score(self):
        # overall value score
        ep = self.process_ep()
        bp = self.process_bp()
        rp = self.process_rp()
        cp = self.process_cp()

        # overall growth score
        er = self.process_er()
        br = self.process_br()
        rr = self.process_rr()
        cr = self.process_cr()

        # value core growth
        score = pd.concat([ep,bp,rp,cp,er,br,rr,cr], axis=1)
        score.loc[:,'ovs'] = score.ep * 0.5 + (score.bp + score.rp + score.cp) / 6
        score.loc[:,'ogs'] = (score.er + score.br + score.rr + score.cr) / 4
        score.loc[:,'vcg'] = score.ogs - score.ovs
        return score

    def process_stock_style(self):
        score = self.calculate_stock_score()
        capsum = self.cap.dropna().sort_values(by='cap',ascending=False).cumsum()
        capt = capsum / capsum.iloc[-1]
        self.big_cap_stock = capt[capt.cap < 0.7].index
        self.mid_cap_stock = capt[np.logical_and(capt.cap  >= 0.7, capt.cap  <= 0.9)].index
        self.small_cap_stock = capt[capt.cap  > 0.9].index
        self.big_vcg = score.loc[score.index.intersection(self.big_cap_stock),'vcg']
        self.mid_vcg = score.loc[score.index.intersection(self.mid_cap_stock),'vcg']
        self.small_vcg = score.loc[score.index.intersection(self.small_cap_stock),'vcg']
        LMT = self.cap.loc[self.mid_cap_stock[0],'cap']
        MST = self.cap.loc[self.mid_cap_stock[-1],'cap']
        res = []
        for stock_id in score.dropna(subset=['vcg']).index:
            y_raw = 100 * (1 + (np.log(self.cap.loc[stock_id,'cap']) - np.log(MST)))/(np.log(LMT) - np.log(MST))
            if y_raw > 200:
                x_raw = 100 * (1 + (score.loc[stock_id,'vcg'] - self.big_vcg.quantile(1 / 3)) / (self.big_vcg.quantile(2 / 3) - self.big_vcg.quantile(1 / 3)))
            elif y_raw < 100:
                x_raw = 100 * (1 + (score.loc[stock_id,'vcg'] - self.small_vcg.quantile(1 / 3)) / (self.small_vcg.quantile(2 / 3) - self.small_vcg.quantile(1 / 3)))
            else:
                x_raw = 100 * (1 + (score.loc[stock_id,'vcg'] - self.mid_vcg.quantile(1 / 3)) / (self.mid_vcg.quantile(2 / 3) - self.mid_vcg.quantile(1 / 3)))
            dic = {
                'stock_id':stock_id,
                'x':x_raw,
                'y':y_raw,
            }
            res.append(dic)
        df = pd.DataFrame(res).set_index('stock_id')
        # 统一股票代码
        res = []
        for i in df.index:
            mkt = i.split('.')[-1]
            if mkt in ['SH','SZ']:
                i = i.split('.')[0] + ' CH Equity'
            elif mkt == 'HK':
                i = i.split('.')[0].lstrip('0')+' HK Equity'
            else:
                i = i.split('.')[0].lstrip('0')+' US Equity'
            res.append(i)
        df.index = res
        return df

    def process_fund_style(self):
        stock_style = self.process_stock_style()
        cols_id = []
        cols_w = []
        for i in range(1,11,1):
            cols_id.append(f'pos_{i}')
            cols_w.append(f'pos_{i}_w')
        res = []
        for fund_id in self.fund_pos.codes:
            _df = self.fund_pos[self.fund_pos.codes == fund_id]
            asset_list = _df[cols_id].values[0].tolist()
            equity_list = [i.split(' ')[-1] for i in asset_list if i is not None]
            if equity_list.count('Equity') < 5:
                continue
            else:
                _df_T = _df.T
                _df_T.columns = [0]
                asset_list = _df_T.loc[cols_id][0].tolist()
                asset_w_list = _df_T.loc[cols_w][0].tolist()
                df_w = pd.DataFrame({'stock_id':asset_list,'stock_w':asset_w_list}).set_index('stock_id')
                df_w = df_w.join(stock_style).dropna()
                x = (df_w.stock_w * df_w.x).sum() / df_w.stock_w.sum()
                y = (df_w.stock_w * df_w.y).sum() / df_w.stock_w.sum()
                dic = {
                    'codes':fund_id,
                    'x':x,
                    'y':y,
                    'datetime':_df.report_date.values[0],
                }
                res.append(dic)
        df_result = pd.DataFrame(res)
        return df_result.dropna()

    def process(self):
        self.result = self.process_fund_style()
        self._data_helper._upload_raw(self.result, QSFundStyleBox.__table__.name)



import json
import os
import datetime
import traceback
from typing import Dict, List, Optional

import pandas as pd
import numpy as np

from .....constant import SourceType
from .....fund.engine.trader import BasicFundTrader
from .....fund.engine.backtest import BasicFundBacktestEngine
from ....fund.basic.basic_data_helper import BasicDataHelper
from ....manager.manager_fund import FundDataManager
from ....view.derived_models import ThirdPartyPortfolioInfo, ThirdPartyPortfolioTrade, ThirdPartyPortfolioIndicator, ThirdPartyPortfolioPositionLatest
from ..derived_data_helper import DerivedDataHelper, DateJSONEncoder
from .common import ThirdPortfolioBase


class DanjuanPortfolioProcessor(ThirdPortfolioBase):

    _FILE_SUFFIX = 'json'
    _INFO_COLUMNS_MAP = {
        'portfolio_id': 'po_id',
        'plan_name': 'po_name',
        'found_date': 'start_date',
        'auto_invest_status': 'can_aip',
    }

    def __init__(self, data_helper: DerivedDataHelper):
        super().__init__()

        self._data_helper = DerivedDataHelper()
        self._params_json: Dict[str, str] = {}
        self._dir_path = f'{os.path.dirname(os.path.abspath(__file__))}/data/danjuan'
        self._trades_df_list = []
        self._nav: Dict[str, pd.DataFrame] = {}
        self._last_position: List[pd.Series] = []

    def init(self):
        # self._fdm = FundDataManager(s3_retriever_activation=True)
        # self._fdm.init(score_pre_calc=True, print_time=True)

        from .....resource.data_store import DataStore
        self._fdm: FundDataManager = DataStore.load_dm()
        self._bdh = BasicDataHelper()
        self._trading_day: pd.DataFrame = FundDataManager.basic_data('get_trading_day_list').drop(columns='_update_time')

        self._list_no_aip = pd.read_csv(f'{self._dir_path}/list_no_aip.txt', header=None, names=['file_name'])
        # self._list_no_aip = self._list_no_aip[self._list_no_aip.file_name.str.split('.', n=1, expand=True)[0].isin(['CSI1003', 'CSI1030', 'CSI1039', 'CSI1046', 'CSI1075', 'CSI1080'])]

    def _calc_basic_info(self, x: pd.Series):
        with open(f'{self._dir_path}/portfolios/{x.file_name}', 'r') as f:
            portfolio_base_info = json.load(f)

        base_info = portfolio_base_info.get('data')
        if base_info is None:
            print(f'[WARNING] Can not find info of {x.file_name}')
            return

        result = {}
        result['po_id'] = base_info.get('plan_code')
        assert result['po_id'] == x.file_name.split('.')[0], 'portfolio id does not match'
        for one in ('plan_name', 'manager_name', 'found_date', 'auto_invest_status'):
            result[one] = base_info.get(one)
        if result['auto_invest_status'] is not None:
            result['auto_invest_status'] = (result['auto_invest_status'] != '0')

        plan_desc = base_info.get('plan_desc')
        if plan_desc is not None:
            for one in plan_desc:
                if one.get('name') == '组合概要':
                    result['rich_desc'] = one.get('plan_desc')
                    break
        op_fund = base_info.get('op_fund')
        if op_fund is not None:
            result['po_desc'] = op_fund.get('recom_desc')
            result['risk_level'] = op_fund.get('risk_tag')
            result['benchmarks'] = op_fund.get('st_code')
        return pd.Series(result)

    def _calc_trades(self, x: pd.Series):
        po_id = x.file_name.split('.')[0]

        fund_weight_list = []
        result_params = {}
        result_params['fund_weight_list'] = []
        with open(f'{self._dir_path}/portfolios/{po_id}_adj.{self._FILE_SUFFIX}', 'r') as f:
            adj_info = json.load(f)

        content = adj_info.get('data')
        if content is None:
            print(f'[WARNING] Invalid content {adj_info}')
        items = content.get('items')
        if items is None:
            print(f'[WARNING] Invalid items {content}')

        for one_item in items:
            assert one_item.get('plan_code') == po_id, 'portfolio id does not match'
            trade_date = datetime.date.fromtimestamp(one_item.get('trade_date') / 1000)
            details = one_item.get('trading_elements')
            if details is None:
                print(f'[WARNING] Invalid details {one_item}')
                continue
            if not details:
                print(f'[INFO] Empty details, do next one')
                continue
            fp_sum = 0
            tp_sum = 0
            result_details = []
            for one in details:
                raw_fund_id = one.get('fd_code')
                fund_id = self._bdh._get_fund_id_from_order_book_id(raw_fund_id, trade_date)
                if fund_id is None:
                    print(f'[WARNING] Can not found fund {raw_fund_id} (po_id){po_id}')
                    fund_id = raw_fund_id + '!0'
                from_percent = one.get('last_portion')
                to_percent = one.get('portion')
                result_details.append({'fund_id': fund_id, 'from_percent': from_percent, 'to_percent': to_percent})
                fp_sum += from_percent
                tp_sum += to_percent
            fp_sum = round(fp_sum, 3)
            tp_sum = round(tp_sum, 3)
            if (fp_sum != 0 and fp_sum != 1) or tp_sum != 1:
                print(f'sum of percent not equals 1 (po_id){po_id} (trade_date){trade_date} (fp_sum){fp_sum} (tp_sum){tp_sum}')
            adjusted_date = trade_date
            # 我们是第二天调仓，因此这里调整为之前的第一个交易日
            adjusted_date = self._trading_day[self._trading_day.datetime < adjusted_date].datetime.array[-1]
            fund_weight_list.append({'po_id': po_id, 'datetime': trade_date, 'details': json.dumps(result_details), 'comment': one_item.get('remark')})
            result_params['fund_weight_list'].append([adjusted_date, [(one['fund_id'], one['to_percent']) for one in result_details]])
        self._trades_df_list.append(pd.DataFrame(fund_weight_list))

        with open(f'{self._dir_path}/portfolios/{x.file_name}', 'r') as f:
            portfolio_base_info = json.load(f)
        base_info = portfolio_base_info.get('data')
        if base_info is None:
            print(f'[WARNING] Can not find info of {x.file_name}')
        else:
            result_params['start_date'] = base_info.get('found_date')
            plan_derived = base_info.get('plan_derived')
            if plan_derived is not None:
                result_params['end_date'] = plan_derived.get('end_date')
            else:
                print(f'[WARNING] Can not find end_date of {x.file_name}')
                result_params['end_date'] = datetime.date.today().isoformat()
        result_params['fund_weight_list'].sort(key=lambda x: x[0])
        self._params_json[po_id] = json.dumps(result_params, indent=4, cls=DateJSONEncoder)

    def _calc_indicators(self, x: pd.Series):
        po_id = x.file_name.split('.')[0]
        params = json.loads(self._params_json.get(po_id))

        # print(po_id)
        # print(params)
        start_date = pd.to_datetime(params.get('start_date'), infer_datetime_format=True).date()
        # 我们是第二天调仓，因此这里调整为之前的第一个交易日
        start_date = self._trading_day[self._trading_day.datetime < start_date].datetime.array[-1]
        # end_date = pd.to_datetime(params.get('end_date'), infer_datetime_format=True).date()

        t = BasicFundTrader()
        b = BasicFundBacktestEngine(data_manager=self._fdm, trader=t, fund_weight_list=params['fund_weight_list'])
        b.init()
        b.run(start_date=start_date, end_date=None, print_time=False)

        last_position: Optional[pd.DataFrame] = b.get_last_position()
        if last_position is None:
            print(f'[WARNING] {po_id} has no last position')
        else:
            self._last_position.append(pd.Series({'po_id': po_id, 'datetime': last_position.index.name, 'po_src': SourceType.danjuan, 'position': last_position.loc[:, 'weight'].to_json()}))

        result = b.get_fund_result()
        if result is None:
            print(f'[WARNING] {po_id} has no fund result')
            return
        # 只保留我们需要的回测结果
        result = {k: v for k, v in result.items() if k in ('market_value', 'last_unit_nav', 'mdd', 'annual_ret', 'mdd_d1', 'mdd_d2', 'total_commission', 'rebalance_times')}
        result['datetime'] = result['market_value'].index.array[-1]
        result['from_setup_ret'] = result['last_unit_nav'] - 1
        resampled = result['market_value'].copy()
        resampled.index = pd.to_datetime(resampled.index)
        resampled_w = resampled.mv.resample('1W').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['vol_by_week'] = resampled_w.dropna().pct_change().std(ddof=1) * np.sqrt(52)

        resampled_m = resampled.mv.resample('1M').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['vol_by_month'] = resampled_m.dropna().pct_change().std(ddof=1) * np.sqrt(12)
        result['vol_by_day'] = resampled.mv.dropna().pct_change().std(ddof=1) * np.sqrt(244)
        result['annual_compounded_ret'] = np.power(result['last_unit_nav'], 1 / (result['market_value'].shape[0] / 244)) - 1
        if result['vol_by_week'] != 0:
            result['sharpe_ratio'] = (result['annual_compounded_ret'] - 0.02) / result['vol_by_week']
        else:
            result['sharpe_ratio'] = np.nan
        # 遍历每个需要计算的时间区间
        for name, value in self._DIM_TYPE.items():
            bd = (result['datetime'] - value).date()
            # 试图寻找bd及其之前的最近一天
            filtered_mv = result['market_value'].loc[result['market_value'].index <= bd, :]
            if filtered_mv.empty:
                continue
            # 计算区间内的return
            result[name] = result['market_value'].mv.array[-1] / filtered_mv.mv.array[-1] - 1

        result['nav'] = result['last_unit_nav']
        result['po_id'] = po_id
        self._nav[po_id] = result['market_value'].mv / result['market_value'].mv.array[0]
        self._nav[po_id].index = self._nav[po_id].index.map(lambda x: x.isoformat())
        # 删掉我们不需要的回测结果
        for one in ('market_value', 'last_unit_nav'):
            del result[one]
        print(f'{po_id} done')
        return pd.Series(result)

    def calc(self):
        basic_info: pd.DataFrame = self._list_no_aip.apply(self._calc_basic_info, axis=1)
        basic_info['po_src'] = SourceType.danjuan
        self._data_helper._upload_derived(basic_info.rename(columns=self._INFO_COLUMNS_MAP), ThirdPartyPortfolioInfo.__table__.name)

        self._list_no_aip.apply(self._calc_trades, axis=1)
        trades = pd.concat(self._trades_df_list)
        trades['po_src'] = SourceType.danjuan
        self._data_helper._upload_derived(trades, ThirdPartyPortfolioTrade.__table__.name)

        indicators: pd.DataFrame = self._list_no_aip.apply(self._calc_indicators, axis=1)
        indicators = indicators[indicators.notna().any(axis=1)]
        indicators['po_src'] = SourceType.danjuan
        self._data_helper._upload_derived(indicators, ThirdPartyPortfolioIndicator.__table__.name)

        last_position = pd.DataFrame(self._last_position)
        self._data_helper._upload_derived(last_position, ThirdPartyPortfolioPositionLatest.__table__.name)

        # nav存cas
        from ....view.cas.third_party_portfolio_nav import ThirdPartyPortfolioNav
        for po_id, mv in self._nav.items():
            ThirdPartyPortfolioNav.create(
                po_id=po_id,
                date=mv.index.to_list(),
                nav=mv.to_list(),
            )

    def process(self, end_date: str) -> List[str]:
        failed_tasks: List[str] = []
        try:
            self.calc_everyday(end_date, [SourceType.danjuan.value])
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('third_party_portfolio_danjuan')
        return failed_tasks


if __name__ == '__main__':
    danjuan_pp = DanjuanPortfolioProcessor(DerivedDataHelper())
    # danjuan_pp.init()
    # danjuan_pp.calc()
    # danjuan_pp.process('20201012')



import os
import ssl
import json
import time
import requests
import multiprocessing.dummy as mp


class QiemanCrawler(object):
    def __init__(self):
        self.portfolio_url_prefix = 'https://qieman.com/pmdj/v1/pomodels/'
        self.result_dir = './'
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.5 Safari/605.1.15',
            'Accept': 'application/json, text/plain, */*',
            'x-sign': '1602564825201B1482E004AA21C91B8F184BB5524C048',
            'Host': 'qieman.com',
            'Referer': 'https://qieman.com/strategies/earl28/detail',
        }

    def setup_proxy(self):
        # 代理隧道验证信息
        proxyUser = "ZPQZ1677448071229353"
        proxyPass = "6pVOQo0ZgMFx"
        # 代理服务器
        proxyHost = "dyn.horocn.com"
        proxyPort = "50000"
        proxyMeta = f"http://{proxyUser}:{proxyPass}@{proxyHost}:{proxyPort}"
        self.proxies = {
            "http": proxyMeta,
            "https": proxyMeta,
        }

    def crawl(self, portfolio_id):
        portfolio_id = f'ZH{str(portfolio_id).zfill(6)}'
        url = f'{self.portfolio_url_prefix}{portfolio_id}'

        retry = 0
        while retry < 2:
            if retry > 0:
                print(f'Retry {retry} times for index {portfolio_id}')

            try:
                # Crawl portfolio description
                # try:
                #     # r = requests.get(url=url, headers=self.headers, proxies=self.proxies, timeout=3)
                #     r = requests.get(url=url, headers=self.headers, timeout=3)
                #     result_json = r.json()
                #     if 'code' in result_json and result_json['code'] == "1600":
                #         print(f'Portfolio {portfolio_id} does not exist!')
                #         return True

                #     result_file = os.path.join(self.result_dir, f'{portfolio_id}.json')
                #     with open(result_file, 'w') as fout:
                #         fout.write(json.dumps(result_json, ensure_ascii=False, indent=4))

                # except json.decoder.JSONDecodeError as e:
                #     print(print(f'Failed to parse portfolio "{r}" for {portfolio_id}! {e}'))
                #     return True

                # Crawl portfolio adjustments
                page_index = 0
                finished = False
                while not finished:
                    url = f'{self.portfolio_url_prefix}{portfolio_id}/adjustments?page={page_index}&size=100&format=openapi&isDesc=true'
                    print(url)
                    # url = f'{self.portfolio_url_prefix}{portfolio_id}/nav-history?'
                    try:
                        # r = requests.get(url=url, headers=self.headers, proxies=self.proxies, timeout=3)
                        r = requests.get(url=url, headers=self.headers, timeout=3)
                        result_json = r.json()
                        finished = result_json['last']

                        result_file = os.path.join(self.result_dir, f'{portfolio_id}_adj{page_index}.json')
                        with open(result_file, 'w') as fout:
                            fout.write(json.dumps(result_json, ensure_ascii=False, indent=4))
                        page_index += 1

                    except json.decoder.JSONDecodeError as e:
                        print(print(f'Failed to parse adjustments "{r}" for {portfolio_id}! {e}'))
                        return True

                print(f'Portfolio {portfolio_id} got')
                return True

            except Exception as e:
                print(f'[WARN] at {portfolio_id}, {e}')
                retry += 1

        print(f'[ERROR] for index {portfolio_id}')
        return False

    def test(self, portfolio_id):
        url = 'http://api.ip.sb/ip'
        r = requests.get(url=url, headers=self.headers, proxies=self.proxies, timeout=3)
        print(r.content.decode())

    def parallel_craw(self, start, end):
        p = mp.Pool(3)
        # p.map(self.test, range(start, end))
        p.map(self.crawl, range(start, end))
        p.close()
        p.join()

    def linear_craw(self, start, end):
        for i in range(start, end):
            # self.test(i)
            self.crawl(i)

    def craw_targets(self, pids):
        for i in pids:
            self.crawl(i)
            # self.test(i)

    def parallel_craw_targets(self, pids):
        p = mp.Pool(3)
        p.map(self.crawl, pids)
        p.close()
        p.join()


if __name__ == '__main__':
    crawler = QiemanCrawler()
    # crawler.setup_proxy()

    start = 17252
    end = 17253

    crawler.linear_craw(start, end)
    # crawler.parallel_craw(start, end)

    # cannot be crawled <= 50000
    # pids = [6536, 6910, 7255, 8508, 10261, 17252, 17255, 14162, 17751, 19522, 25902, 25904, 34050, 35300, 35567]
    # pids = [16]
    # crawler.parallel_craw_targets(pids)



import pandas as pd
import datetime
from pandas.tseries.offsets import DateOffset
from ...util.singleton import Singleton
from ..wrapper.mysql import ViewDatabaseConnector
from ..view.view_models import *


class ViewDataApi(metaclass=Singleton):
    @staticmethod
    def get_fund_daily_collection():
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(FundDailyCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=FundDailyCollection.trans_columns())
                df['fund_type'] = df['基金大类类型']
                return df
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_daily_index_collection():
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(IndexDailyCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=IndexDailyCollection.trans_columns())
                return df
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_daily_fund_automatic_investment():
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(AutomaticInvestmentCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=AutomaticInvestmentCollection.trans_columns())
                return df
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_daily_fund_daily_result(fund_id_list = []):
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundDailyCollection.fund_id,
                    FundDailyCollection.return_score,
                    FundDailyCollection.robust_score,
                    FundDailyCollection.risk_score,
                    FundDailyCollection.timing_score,
                    FundDailyCollection.selection_score,
                    FundDailyCollection.team_score,
                    FundDailyCollection.total_score,
                ).filter(
                    FundDailyCollection.fund_id.in_(fund_id_list)
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=FundDailyCollection.trans_columns())
                return df
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_recent_stock_ipo():
        def _get_relate_date_df(df, dt):
            _df = df[df.上市日期 > dt]
            _df_normal = _df[_df.交易板块.isin(['深交所中小板','上交所主板'])]
            _df_gem = _df[_df.交易板块 == '深交所创业板']
            _df_star = _df[_df.交易板块 == '上交所科创板']
            res = {
                'data':_df,
                '普通新股只数':_df_normal.shape[0],
                '普通新股平均连板天数':_df_normal.连续涨停.mean(),
                '普通新股平均每签获利':_df_normal.每签获利.mean(),
                '创业板新股只数':_df_gem.shape[0],
                '创业板新股首日破发数':_df_gem[_df_gem.首日涨幅 < 0].shape[0],
                '创业板平均首日涨幅':_df_gem.首日涨幅.mean(),
                '科创新股只数':_df_star.shape[0],
                '科创新股首日破发数':_df_star[_df_star.首日涨幅 < 0].shape[0],
                '科创新股平均每日涨幅':_df_star.首日涨幅.mean(),
            }
            return res

        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(StockIpoCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=StockIpoCollection.trans_columns()).drop(columns='_update_time')
                today = datetime.datetime.now().date()
                date_3m_ago = (today - DateOffset(months=3)).date()
                date_6m_ago = (today - DateOffset(months=6)).date()
                date_1y_ago = (today - DateOffset(years=1)).date()
                result = {
                    '近三月':_get_relate_date_df(df, date_3m_ago),
                    '近六月':_get_relate_date_df(df, date_6m_ago),
                    '近一年':_get_relate_date_df(df, date_1y_ago),
                }
                return result
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_recent_conv_bond_ipo():
        def _get_relate_date_df(df, dt):
            _df = df[df.上市日期 > dt]
            res = {
                'data':_df,
                '上市数量':_df.shape[0],
                '首日破发数':_df[_df.首日涨幅 < 0].shape[0],
                '首日破发率':_df[_df.首日涨幅 < 0].shape[0] / _df.shape[0],
                '每签最大收益':_df.每签获利.max(),
                '每签最大亏损':abs(min(_df.每签获利.min(), 0)),
                '单账户累计收益':_df.单账户收益.sum(),
            }
            return res
        
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(ConvBondIpoCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=ConvBondIpoCollection.trans_columns()).drop(columns='_update_time')
                today = datetime.datetime.now().date()
                date_3m_ago = (today - DateOffset(months=3)).date()
                date_6m_ago = (today - DateOffset(months=6)).date()
                date_1y_ago = (today - DateOffset(years=1)).date()
                result = {
                    '近三月':_get_relate_date_df(df, date_3m_ago),
                    '近六月':_get_relate_date_df(df, date_6m_ago),
                    '近一年':_get_relate_date_df(df, date_1y_ago),
                }
                return result
            except Exception as e:
                print(e)
                return False



import datetime
import traceback
import json
import numpy as np
import pandas as pd
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from .basic_data_helper import BasicDataHelper
from ...view.basic_models import OverseaFundInfo, OverseaFundNav

class OtherBasicData:

    def __init__(self, data_helper: BasicDataHelper):
        self._data_helper = data_helper
        self._raw_data_api = RawDataApi()
        self._basic_data_api = BasicDataApi()

    def make_fund_id_func(self):
        return lambda x : 'os_' + x

    def oversea_fund_nav(self):
        df = self._raw_data_api.get_oversea_fund_nav()
        df['codes'] = df.codes.map(self.make_fund_id_func)
        df = df.drop(columns=['_update_time'])
        self._data_helper._upload_basic(df, OverseaFundNav.__table__.name)

    def oversea_fund_info(self):
        df = self._raw_data_api.get_over_sea_fund_info()
        df['codes'] = df.codes.map(self.make_fund_id_func)
        df = df.drop(columns=['_update_time'])
        self._data_helper._upload_basic(df, OverseaFundInfo.__table__.name)



import pandas as pd

# from typing import Optional

from ...constant import StockFactorType
from .base import Factor
from .utils import get_factor_from_stock_data, calc_ttm


class BasicFactor(Factor):
    # 所有基础因子的基类
    def __init__(self, name):
        super().__init__(f'basic/{name}', StockFactorType.BASIC)


class StockTradableFactor(BasicFactor):
    # 交易状态
    def __init__(self):
        super().__init__('stock_tradable')

    def calc(self):
        tp = get_factor_from_stock_data('get_em_stock_post_price', columns=['trade_status'])
        tp[tp == '正常交易'] = 1
        tp[tp == '复牌'] = 1
        tp[tp == '盘中停牌'] = 1
        tp[tp == '停牌半小时'] = 1
        tp[tp == '停牌一小时'] = 1
        tp[tp == '停牌半天'] = 1
        tp[tp != 1] = 0
        self._factor = tp


class StockPriceFactor(BasicFactor):
    # 未复权股价
    def __init__(self):
        super().__init__('stock_price')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_price', columns=['close'])


class StockPostPriceFactor(BasicFactor):
    # 后复权股价
    def __init__(self):
        super().__init__('stock_post_price')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_post_price', columns=['close'])
        self._factor_return = self._factor.pct_change(fill_method=None)


class StockOpenPriceFactor(BasicFactor):
    # 开盘价
    def __init__(self):
        super().__init__('stock_open_price')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_post_price', columns=['open'])


class StockHighPriceFactor(BasicFactor):
    # 最高价
    def __init__(self):
        super().__init__('stock_high_price')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_post_price', columns=['high'])


class StockLowPriceFactor(BasicFactor):
    # 最低价
    def __init__(self):
        super().__init__('stock_low_price')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_post_price', columns=['low'])


class StockVolumeFactor(BasicFactor):
    # 成交量
    def __init__(self):
        super().__init__('stock_volume')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_post_price', columns=['volume'])
        self._factor = self._factor.ffill()


class StockVWAPFactor(BasicFactor):
    # VWAP
    def __init__(self):
        super().__init__('stock_vwap')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_post_price', columns=['average'])
        self._factor = self._factor.ffill()


class TotalShareFactor(BasicFactor):
    # 总股本
    def __init__(self):
        super().__init__('total_share')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_daily_info', columns=['total_share'])


class CirculationShareFactor(BasicFactor):
    # 流通股本
    def __init__(self):
        super().__init__('circ_share')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_daily_info', columns=['liq_share'])


class MarketValueFactor(BasicFactor):
    # 总市值
    def __init__(self):
        super().__init__('market_value')
        self._deps.add(TotalShareFactor())
        self._deps.add(StockPostPriceFactor())

    def calc(self):
        # 总市值 = 总股本 * 每日后复权股价
        self._factor = TotalShareFactor().get() * StockPostPriceFactor().get()


class PreferredStockFactor(BasicFactor):
    # 优先股
    def __init__(self):
        super().__init__('preferred_stock')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_196'])


class AdjNetAssetAvgFactor(BasicFactor):
    # 经优先股调整的净资产最近四期均值
    def __init__(self):
        super().__init__('adj_net_asset_avg')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_141'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean()) - \
            get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_196'], processor_after_pivot=lambda x: x.fillna(0).rolling(4, min_periods=1).mean())


class TotalAssetAvgFactor(BasicFactor):
    # 总资产最近四期均值
    def __init__(self):
        super().__init__('total_asset_avg')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_74'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean())


class TotalDebtAvgFactor(BasicFactor):
    # 总负债最近四期均值
    def __init__(self):
        super().__init__('total_debt_avg')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_128'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean())


class NetAssetFactor(BasicFactor):
    # 净资产
    def __init__(self):
        super().__init__('net_asset')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_141'])


class FixAssetAvgFactor(BasicFactor):
    # 固定资产最近四期均值
    def __init__(self):
        super().__init__('fix_asset_avg')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_31'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean())


class NetProfitFactor(BasicFactor):
    # 净利润
    def __init__(self):
        super().__init__('net_profit')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['income_statement_60'])


class NetProfitTTMFactor(BasicFactor):
    # 净利润TTM
    def __init__(self):
        super().__init__('net_profit_ttm')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['np_ttmrp'])


class IncomeFactor(BasicFactor):
    # 营业收入
    def __init__(self):
        super().__init__('income')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['income_statement_9'])


class IncomeTTMFactor(BasicFactor):
    # 营业收入TTM
    def __init__(self):
        super().__init__('income_ttm')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['or_ttmr'])


class ExpenseTTMFactor(BasicFactor):
    # 营业支出TTM
    def __init__(self):
        super().__init__('expense_ttm')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['expense_ttmr'])


class AdjNetProfitFactor(BasicFactor):
    # 归属于上市公司股东的扣除非经常性损益后的净利润(调整前)
    def __init__(self):
        super().__init__('adj_net_profit')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['deducted_income_ba'])


class TotalRevenueTTMFactor(BasicFactor):
    # 营业总收入TTM
    def __init__(self):
        super().__init__('total_revenue_ttm')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['income_statement_83'], processor_after_pivot=lambda x: x.apply(calc_ttm, axis=1, result_type='broadcast', whole_df=x))


class DividendYearlyFactor(BasicFactor):
    # 年度累计分红总额
    def __init__(self):
        super().__init__('dividend_yearly')

    def calc(self):
        def _resample_to_date(df):
            # 年份变为日期(xxxx-01-01)
            df.index = pd.to_datetime(df.index, format='%Y')
            # 01-01变为12-31
            df = df.resample('1Y').mean()
            df.index = df.index.date
            return df

        self._factor = get_factor_from_stock_data('get_em_stock_yearly', columns=['div_annu_accum'], index='year', processor_after_pivot=_resample_to_date).fillna(0)


class EBITDAFactor(BasicFactor):
    # EBITDA
    def __init__(self):
        super().__init__('ebitda')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['ebitda'])


class EBITDATTMFactor(BasicFactor):
    # EBITDA TTM
    def __init__(self):
        super().__init__('ebitda_ttm')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['ebitda'], processor_after_pivot=lambda x: x.apply(calc_ttm, axis=1, result_type='broadcast', whole_df=x))
        # data_to_calc_ebitda_manually = get_factor_from_stock_data('get_em_stock_fin_fac', columns=[
        #     'income_statement_61',
        #     'income_statement_56',
        #     'cashflow_statement_87',
        #     'cashflow_statement_88',
        #     'cashflow_statement_89',
        #     'cashflow_statement_70',
        # ])
        # self._factor = ebitda.fillna(data_to_calc_ebitda_manually.sum(axis=1, level=0, min_count=1))


class EntValueFactor(BasicFactor):
    # 企业价值
    def __init__(self):
        super().__init__('ent_value')
        self._deps.add(TotalDebtAvgFactor())
        self._deps.add(MarketValueFactor())

    def calc(self):
        # def _resample_to_yearly(df):
        #     df.index = pd.to_datetime(df.index, infer_datetime_format=True)
        #     # 只保留年报数据
        #     df = df.resample('1Y').last()
        #     df.index = df.index.date
        #     return df

        # self._factor = get_factor_from_stock_data('get_em_daily_info', columns=['ev_without_cash'])
        # 目前没有东财的数据，我们自己计算一下
        # total_debt = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_128'], processor_after_pivot=_resample_to_yearly)
        total_debt = TotalDebtAvgFactor().get()
        monetary_fund = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_9'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean())
        self._factor = MarketValueFactor().get() + total_debt - monetary_fund


class CashFlowFactor(BasicFactor):
    # 经营性现金流
    def __init__(self):
        super().__init__('cash_flow')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['cashflow_statement_39'])


class CashFlowTTMFactor(BasicFactor):
    # 经营性现金流TTM
    def __init__(self):
        super().__init__('cash_flow_ttm')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['cashflow_statement_39'], processor_after_pivot=lambda x: x.apply(calc_ttm, axis=1, result_type='broadcast', whole_df=x))


class GrossProfitFactor(BasicFactor):
    # 毛利TTM
    def __init__(self):
        super().__init__('gross_profit')
        self._deps.add(IncomeTTMFactor())
        self._deps.add(ExpenseTTMFactor())

    def calc(self):
        df = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['gross_margin_ttmr'])
        # 由于银行等金融类企业的财报里没有毛利润, 这里我们使用毛利润TTM = 营业收入TTM - 营业支出TTM来计算
        gross_profit_manullay = IncomeTTMFactor().get() - ExpenseTTMFactor().get()
        self._factor = df.fillna(gross_profit_manullay)


class FloatAssetAvgFactor(BasicFactor):
    # 流动资产最近四期均值
    def __init__(self):
        super().__init__('float_asset_avg')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_25'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean())


class FloatDebtAvgFactor(BasicFactor):
    # 流动负债最近四期均值
    def __init__(self):
        super().__init__('float_debt_avg')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_93'], processor_after_pivot=lambda x: x.rolling(4, min_periods=1).mean())


class MonetaryFundFactor(BasicFactor):
    # 货币资金
    def __init__(self):
        super().__init__('monetary_fund')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_fin_fac', columns=['balance_statement_9'])


class TurnoverFactor(BasicFactor):
    # 换手率
    def __init__(self):
        super().__init__('turnover')

    def calc(self):
        self._factor = get_factor_from_stock_data('get_em_stock_price', columns=['turn'])


class PredictEPSFactor(BasicFactor):
    # 分析师一致预测EPS
    pass



from ...manager.manager_fund import FundDataManager
from ...manager.score import FundScoreManager
from ...struct import AssetWeight, FundScoreHelper
from ...wrapper.mysql import DerivedDatabaseConnector
from ...view.derived_models import FundScore
import pandas as pd
import traceback
import datetime

class FundScoreProcessor(object):

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def init(self, start_time, end_time):
        self._dm = FundDataManager(start_time=start_time, end_time=end_time, score_manager=FundScoreManager())
        self._dm.init()
        self.index_tag = FundScoreHelper()
        self.index_list = list(AssetWeight.__dataclass_fields__.keys())

        start_date_dt = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        end_date_dt = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.date_list = self._dm.dts.fund_indicator.datetime
        self.date_list = self.date_list[(self.date_list >= start_date_dt) & (self.date_list <= end_date_dt)]
        self.date_list = sorted(list(set(self.date_list)))
        self.desc_dic = self._dm.dts.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict('index')

    def calculate_score(self):
        res = []
        for dt in self.date_list:
            for index_id in self.index_list:
                try:
                    res_i, _ = self._dm.get_fund_score(dt, index_id)
                    if res_i:
                        for fund_id, score_i in res_i.items():
                            dic = {
                                'fund_id':fund_id,
                                'score':score_i,
                                'datetime':dt,
                                'index_id':index_id
                            }
                            res.append(dic)
                except:
                    continue
        score_df = pd.DataFrame(res)
        score_df['tag_name'] = score_df['index_id'].map(lambda x: self.index_tag.get(x))
        score_df.drop(['index_id'], axis=1, inplace=True)
        score_df['is_full'] = 1
        score_df['tag_type'] = 1
        score_df['tag_method'] = 'm0'
        score_df['score_method'] = 9
        score_df['desc_name'] = score_df['fund_id'].map(lambda x: self.desc_dic[x]['desc_name'])
        self._data_helper._upload_derived(score_df, FundScore.__table__.name)

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            self.init(start_date, end_date)
            self.calculate_score()
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_score')

        return failed_tasks



import argparse
import subprocess
import sys

HELP_MSG = '''
ifa <command> [<args>]

The most commonly used commands are:
    run [seq] [config_file_path]                run IFA service in foreground
'''

class IfaEntrance(object):

    def __init__(self):
        parser = argparse.ArgumentParser(description='IFA Command Tool', usage=HELP_MSG)
        parser.add_argument('command', help='Subcommand to run')
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print(f'Unrecognized command {args.command}\n')
            parser.print_help()
            exit(1)

        # ------ parser start ------
        self.main_parser = argparse.ArgumentParser(description='gun run')
        self.main_parser.add_argument('-s', '--msg_seq', help='Message sequence from which to start monitoring')
        self.main_parser.add_argument('-c', '--cfg', help='config file path')
        # ------ parser end ------

        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def run(self):
        from .ifa_runner import IfaRunner
        from .util.config import Configurator

        seq = None
        if len(sys.argv) > 1:
            args = self.main_parser.parse_args(sys.argv[2:])
            print(f'seq: {args.msg_seq}, cfg: {args.cfg}')

            if args.msg_seq:
                seq = int(args.msg_seq)

            if args.cfg:
                Configurator(args.cfg)

        runner = IfaRunner(seq)
        runner.run()
    
    # TODO: Implement start method, to run IFA service in background
    def start(self):
        raise Exception('"start" method not implemented, please use "run" instead.')

if __name__ == "__main__":
    entrance = IfaEntrance()
